from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor
import numpy as np


def _load_required_modules():
    try:
        import exdir, quantities
    except ModuleNotFoundError:
        raise ModuleNotFoundError("To use the ExdirExtractors run:\n\n"
                                  "pip install exdir\n\n")
    return exdir, quantities


class ExdirRecordingExtractor(RecordingExtractor):
    def __init__(self, exdir_file):
        exdir, pq = _load_required_modules()

        RecordingExtractor.__init__(self)
        self._exdir_file = exdir_file
        exdir_group = exdir.File(exdir_file)

        self._recordings = exdir_group['acquisition']['timeseries']
        self._samplerate = self._recordings.attrs['sample_rate']

        self._num_channels = self._recordings.shape[0]
        self._num_timepoints = self._recordings.shape[1]

    def getChannelIds(self):
        return list(range(self._num_channels))

    def getNumFrames(self):
        return self._num_timepoints

    def getSamplingFrequency(self):
        return self._samplerate

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = self.getChannelIds()
        return self._recordings[channel_ids, start_frame:end_frame]

    @staticmethod
    def writeRecording(recording, exdir_file, lfp=False, mua=False):
        exdir, pq = _load_required_modules()

        channel_ids = recording.getChannelIds()
        M = len(channel_ids)
        N = recording.getNumFrames()
        raw = recording.getTraces()
        exdir_group = exdir.File(exdir_file, plugins=exdir.plugins.quantities)

        if not lfp and not mua:
            timeseries = exdir_group.require_group('acquisition').require_dataset('timeseries', data=raw)
            timeseries.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
            return
        elif lfp:
            ephys = exdir_group.require_group('processing').require_group('electrophysiology')
            if 'group' in recording.getChannelPropertyNames():
                channel_groups = np.unique([recording.getChannelProperty(ch, 'group')
                                            for ch in recording.getChannelIds()])
            else:
                channel_groups  =[0]

            if len(channel_groups) == 1:
                chan = 0
                ch_group = ephys.require_group('Channel_group_' + str(chan))
                lfp_group = ch_group.require_group('LFP')
                ch_group.attrs['electrode_group_id'] = chan
                ch_group.attrs['electrode_identities'] = np.arange(len(recording.getChannelIds()))
                ch_group.attrs['electrode_idx'] = np.arange(len(recording.getChannelIds()))
                ch_group.attrs['start_time'] = 0 * pq.s
                ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                              float(recording.getSamplingFrequency()) * pq.s
                for i_c, ch in enumerate(recording.getChannelIds()):
                    ts_group = lfp_group.require_group('LFP_timeseries_' + str(ch))
                    ts_group.attrs['electrode_group_id'] = chan
                    ts_group.attrs['electrode_identity'] = ch
                    ts_group.attrs['num_samples'] = recording.getNumFrames()
                    ts_group.attrs['electrode_idx'] = i_c
                    ts_group.attrs['start_time'] = 0 * pq.s
                    ts_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                  float(recording.getSamplingFrequency()) * pq.s
                    ts_group.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                    data = ts_group.require_dataset('data', data=recording.getTraces(channel_ids=[ch]))
                    data.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                    data.attrs['unit'] = pq.uV
            else:
                channel_groups = np.unique([recording.getChannelProperty(ch, 'group')
                                            for ch in recording.getChannelIds()])
                for chan in channel_groups:
                    ch_group = ephys.require_group('Channel_group_' + str(chan))
                    lfp_group = ch_group.require_group('LFP')
                    ch_group.attrs['electrode_group_id'] = chan
                    ch_group.attrs['electrode_identities'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                               if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['electrode_idx'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                        if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['start_time'] = 0 * pq.s
                    ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                   float(recording.getSamplingFrequency()) * pq.s
                    for i_c, ch in enumerate(recording.getChannelIds()):
                        if recording.getChannelProperty(ch, 'group') == chan:
                            ts_group = lfp_group.require_group('LFP_timeseries_'+str(ch))
                            ts_group.attrs['electrode_group_id'] = chan
                            ts_group.attrs['electrode_identity'] = ch
                            ts_group.attrs['num_samples'] = recording.getNumFrames()
                            ts_group.attrs['electrode_idx'] = i_c
                            ts_group.attrs['start_time'] = 0 * pq.s
                            ts_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                          float(recording.getSamplingFrequency()) * pq.s
                            ts_group.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                            data = ts_group.require_dataset('data', data=recording.getTraces(channel_ids=[ch]))
                            data.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                            data.attrs['unit'] = pq.uV
            return
        elif mua:
            ephys = exdir_group.require_group('processing').require_group('electrophysiology')
            if 'group' in recording.getChannelPropertyNames():
                channel_groups = np.unique([recording.getChannelProperty(ch, 'group')
                                            for ch in recording.getChannelIds()])
            else:
                channel_groups  =[0]

            if len(channel_groups) == 1:
                chan = 0
                ch_group = ephys.require_group('Channel_group_' + str(chan))
                mua_group = ch_group.require_group('MUA')
                ch_group.attrs['electrode_group_id'] = chan
                ch_group.attrs['electrode_identities'] = np.arange(len(recording.getChannelIds()))
                ch_group.attrs['electrode_idx'] = np.arange(len(recording.getChannelIds()))
                ch_group.attrs['start_time'] = 0 * pq.s
                ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                              float(recording.getSamplingFrequency()) * pq.s
                for i_c, ch in enumerate(recording.getChannelIds()):
                    ts_group = mua_group.require_group('MUA_timeseries_' + str(ch))
                    ts_group.attrs['electrode_group_id'] = chan
                    ts_group.attrs['electrode_identity'] = ch
                    ts_group.attrs['num_samples'] = recording.getNumFrames()
                    ts_group.attrs['electrode_idx'] = i_c
                    ts_group.attrs['start_time'] = 0 * pq.s
                    ts_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                  float(recording.getSamplingFrequency()) * pq.s
                    ts_group.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                    data = ts_group.require_dataset('data', data=recording.getTraces(channel_ids=[ch]))
                    data.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                    data.attrs['unit'] = pq.uV
            else:
                channel_groups = np.unique([recording.getChannelProperty(ch, 'group')
                                            for ch in recording.getChannelIds()])
                for chan in channel_groups:
                    ch_group = ephys.require_group('Channel_group_' + str(chan))
                    mua_group = ch_group.require_group('MUA')
                    ch_group.attrs['electrode_group_id'] = chan
                    ch_group.attrs['electrode_identities'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                               if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['electrode_idx'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                        if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['start_time'] = 0 * pq.s
                    ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                   float(recording.getSamplingFrequency()) * pq.s
                    for i_c, ch in enumerate(recording.getChannelIds()):
                        if recording.getChannelProperty(ch, 'group') == chan:
                            ts_group = mua_group.require_group('MUA_timeseries_'+str(ch))
                            ts_group.attrs['electrode_group_id'] = chan
                            ts_group.attrs['electrode_identity'] = ch
                            ts_group.attrs['num_samples'] = recording.getNumFrames()
                            ts_group.attrs['electrode_idx'] = i_c
                            ts_group.attrs['start_time'] = 0 * pq.s
                            ts_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                          float(recording.getSamplingFrequency()) * pq.s
                            ts_group.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                            data = ts_group.require_dataset('data', data=recording.getTraces(channel_ids=[ch]))
                            data.attrs['sample_rate'] = recording.getSamplingFrequency() * pq.Hz
                            data.attrs['unit'] = pq.uV



class ExdirSortingExtractor(SortingExtractor):
    def __init__(self, exdir_file, sample_rate=None):
        exdir, pq = _load_required_modules()

        SortingExtractor.__init__(self)
        self._exdir_file = exdir_file
        exdir_group = exdir.File(exdir_file, plugins=exdir.plugins.quantities)

        if 'acquisition' in exdir_group.keys():
            if 'timeseries' in exdir_group['acquisition'].keys():
                sample_rate = exdir_group['acquisition']['timeseries'].attrs['sample_rate']
        else:
            if sample_rate is None:
                raise Exception("Provide 'sample_rate' argument (Hz)")
            else:
                sample_rate = sample_rate * pq.Hz

        electrophysiology = exdir_group['processing']['electrophysiology']
        self._unit_ids = []
        current_unit = 1
        self._spike_trains = []
        for chan_name, channel in electrophysiology.items():
            group = int(chan_name.split('_')[-1])
            for units, unit_times in channel['UnitTimes'].items():
                self._unit_ids.append(current_unit)
                self._spike_trains.append((unit_times['times'].data.rescale('s')*sample_rate).magnitude)
                self.setUnitProperty(current_unit, 'group', group)
                current_unit += 1

    def getUnitIds(self):
        return self._unit_ids

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._spike_trains[unit_id]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return np.rint(times[inds]).astype(int)

    @staticmethod
    def writeSorting(sorting, exdir_file, recording=None, sample_rate=None):
        exdir, pq = _load_required_modules()

        if sample_rate is None and recording is None:
            raise Exception("Provide 'sample_rate' argument (Hz)")
        else:
            if recording is None:
                sample_rate = sample_rate * pq.Hz
            else:
                sample_rate = recording.getSamplingFrequency() * pq.Hz

        exdir_group = exdir.File(exdir_file, plugins=exdir.plugins.quantities)
        ephys = exdir_group.require_group('processing').require_group('electrophysiology')

        if 'group' in recording.getChannelPropertyNames():
            channel_groups = np.unique([sorting.getUnitProperty(unit, 'group') for unit in sorting.getUnitIds()])
        else:
            channel_groups = [0]

        if len(channel_groups) == 1:
            chan = 0
            print("Single group: ", chan)
            ch_group = ephys.require_group('Channel_group_' + str(chan))
            unittimes = ch_group.require_group('UnitTimes')
            eventwaveform = ch_group.require_group('EventWaveform')
            if recording is not None:
                ch_group.attrs['electrode_group_id'] = chan
                ch_group.attrs['electrode_identities'] = np.array([])
                ch_group.attrs['electrode_idx'] = np.arange(len(recording.getChannelIds()))
                ch_group.attrs['start_time'] = 0 * pq.s
                ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                              float(recording.getSamplingFrequency()) * pq.s
                unittimes.attrs['electrode_group_id'] = chan
                unittimes.attrs['electrode_identities'] = np.array([])
                unittimes.attrs['electrode_idx'] = np.array(recording.getChannelIds())
                unittimes.attrs['start_time'] = 0 * pq.s
                unittimes.attrs['stop_time'] = recording.getNumFrames() / \
                                               float(recording.getSamplingFrequency()) * pq.s
            nums = np.array([])
            timestamps = np.array([])
            waveforms = np.array([])
            for unit in sorting.getUnitIds():
                unit_group = unittimes.require_group(str(unit))
                unit_group.require_dataset('times',
                                           data=(sorting.getUnitSpikeTrain(unit).astype(float)
                                                 / sample_rate).rescale('s'))
                unit_group.attrs['cluster_group'] = 'unsorted'
                unit_group.attrs['group_id'] = chan
                unit_group.attrs['name'] = 'unit #' + str(unit)

                timestamps = np.concatenate((timestamps, (sorting.getUnitSpikeTrain(unit).astype(float)
                                                          / sample_rate).rescale('s')))
                nums = np.concatenate((nums, [unit] * len(sorting.getUnitSpikeTrain(unit))))

                if 'waveforms' in sorting.getUnitSpikeFeatureNames(unit):
                    if len(waveforms) == 0:
                        waveforms = sorting.getUnitSpikeFeatures(unit, 'waveforms')
                    else:
                        waveforms = np.vstack((waveforms, sorting.getUnitSpikeFeatures(unit, 'waveforms')))

            print("Saving eventwaveforms and clustering")
            if 'waveforms' in sorting.getUnitSpikeFeatureNames():
                waveform_ts = eventwaveform.require_group('waveform_timeseries')
                data = waveform_ts.require_dataset('data', data=waveforms)
                data.attrs['num_samples'] = len(waveforms)
                data.attrs['sample_rate'] = sample_rate
                data.attrs['unit'] = pq.dimensionless
                times = waveform_ts.require_dataset('timestamps', data=timestamps)
                times.attrs['num_samples'] = len(timestamps)
                times.attrs['unit'] = pq.s
                waveform_ts.attrs['electrode_group_id'] = chan
                if recording is not None:
                    waveform_ts.attrs['electrode_identities'] = np.array([])
                    waveform_ts.attrs['electrode_idx'] = np.arange(len(recording.getChannelIds()))
                    waveform_ts.attrs['start_time'] = 0 * pq.s
                    waveform_ts.attrs['stop_time'] = recording.getNumFrames() / \
                                                     float(recording.getSamplingFrequency()) * pq.s
                    waveform_ts.attrs['sample_rate'] = sample_rate
                    waveform_ts.attrs['sample_length'] = waveforms.shape[1]
                    waveform_ts.attrs['num_samples'] = len(waveforms)
            clustering = ephys.require_group('Channel_group_' + str(chan)).require_group('Clustering')
            ts = clustering.require_dataset('timestamps', data=timestamps * pq.s)
            ts.attrs['num_samples'] = len(timestamps)
            ts.attrs['unit'] = pq.s
            ns = clustering.require_dataset('nums', data=nums)
            ns.attrs['num_samples'] = len(nums)
            cn = clustering.require_dataset('cluster_nums', data=np.array(sorting.getUnitIds()))
            cn.attrs['num_samples'] = len(sorting.getUnitIds())
        else:
            channel_groups = np.unique([sorting.getUnitProperty(unit, 'group') for unit in sorting.getUnitIds()])
            for chan in channel_groups:
                print("Group: ", chan)
                ch_group = ephys.require_group('Channel_group_' + str(chan))
                unittimes = ch_group.require_group('UnitTimes')
                eventwaveform = ch_group.require_group('EventWaveform')
                if recording is not None:
                    unittimes.attrs['electrode_group_id'] = chan
                    unittimes.attrs['electrode_identities'] = np.array([])
                    unittimes.attrs['electrode_idx'] = np.array([ch for i_c, ch in enumerate(recording.getChannelIds())
                                                                 if recording.getChannelProperty(ch, 'group') == chan])
                    unittimes.attrs['start_time'] = 0 * pq.s
                    unittimes.attrs['stop_time'] = recording.getNumFrames() / \
                                                  float(recording.getSamplingFrequency()) * pq.s
                    ch_group.attrs['electrode_group_id'] = chan
                    ch_group.attrs['electrode_identities'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                                       if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['electrode_idx'] = np.array([i_c for i_c, ch in enumerate(recording.getChannelIds())
                                                       if recording.getChannelProperty(ch, 'group') == chan])
                    ch_group.attrs['start_time'] = 0 * pq.s
                    ch_group.attrs['stop_time'] = recording.getNumFrames() / \
                                                  float(recording.getSamplingFrequency()) * pq.s
                nums = np.array([])
                timestamps = np.array([])
                waveforms = np.array([])
                for unit in sorting.getUnitIds():
                    if sorting.getUnitProperty(unit, 'group') == chan:
                        print("Unit: ", unit)
                        unit_group = unittimes.require_group(str(unit))
                        unit_group.require_dataset('times',
                                                   data=(sorting.getUnitSpikeTrain(unit).astype(float)
                                                         / sample_rate).rescale('s'))
                        unit_group.attrs['cluster_group'] = 'unsorted'
                        unit_group.attrs['group_id'] = chan
                        unit_group.attrs['name'] = 'unit #' + str(unit)

                        timestamps = np.concatenate((timestamps, (sorting.getUnitSpikeTrain(unit).astype(float)
                                                         / sample_rate).rescale('s')))
                        nums = np.concatenate((nums, [unit]*len(sorting.getUnitSpikeTrain(unit))))

                    if 'waveforms' in sorting.getUnitSpikeFeatureNames(unit):
                        if len(waveforms) == 0:
                            waveforms = sorting.getUnitSpikeFeatures(unit, 'waveforms')
                        else:
                            waveforms = np.vstack((waveforms, sorting.getUnitSpikeFeatures(unit, 'waveforms')))
                print("Saving eventwaveforms and clustering")
                if 'waveforms' in sorting.getUnitSpikeFeatureNames():
                    waveform_ts = eventwaveform.require_group('waveform_timeseries')
                    data = waveform_ts.require_dataset('data', data=waveforms)
                    data.attrs['num_samples'] = len(waveforms)
                    data.attrs['sample_rate'] = sample_rate
                    data.attrs['unit'] = pq.dimensionless
                    times = waveform_ts.require_dataset('timestamps', data=timestamps)
                    times.attrs['num_samples'] = len(timestamps)
                    times.attrs['unit'] = pq.s
                    waveform_ts.attrs['electrode_group_id'] = chan
                    if recording is not None:
                        waveform_ts.attrs['electrode_identities'] = np.array([])
                        waveform_ts.attrs['electrode_idx'] = np.array([ch for i_c, ch in
                                                                       enumerate(recording.getChannelIds())
                                                                       if recording.getChannelProperty(ch, 'group') == chan])
                        waveform_ts.attrs['start_time'] = 0 * pq.s
                        waveform_ts.attrs['stop_time'] = recording.getNumFrames() / \
                                                       float(recording.getSamplingFrequency()) * pq.s
                        waveform_ts.attrs['sample_rate'] = sample_rate
                        waveform_ts.attrs['sample_length'] = waveforms.shape[1]
                        waveform_ts.attrs['num_samples'] = len(waveforms)
                clustering = ephys.require_group('Channel_group_' + str(chan)).require_group('Clustering')
                ts = clustering.require_dataset('timestamps', data=timestamps*pq.s)
                ts.attrs['num_samples'] = len(timestamps)
                ts.attrs['unit'] = pq.s
                ns = clustering.require_dataset('nums', data=nums)
                ns.attrs['num_samples'] = len(nums)
                cn = clustering.require_dataset('cluster_nums', data=np.array(sorting.getUnitIds()))
                cn.attrs['num_samples'] = len(sorting.getUnitIds())


from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor

import quantities as pq
import numpy as np
from pathlib import Path
import h5py
import yaml, json
import neo


class MEArecRecordingExtractor(RecordingExtractor):
    def __init__(self, recording_path=None):
        RecordingExtractor.__init__(self)
        self._recording_path = recording_path
        self._fs = None
        self._positions = None
        self._recordings = None
        self._initialize()

    def _initialize(self):
        rec_dict, info = load_recordings(recordings=self._recording_path)

        self._fs = info['recordings']['fs']
        self._recordings = rec_dict['recordings']
        for chan, pos in enumerate(rec_dict['channel_positions']):
            self.setChannelProperty(chan, 'location', pos)

    def getChannelIds(self):
        if self._recordings is None:
            self._initialize()
        return list(range(self._recordings.shape[0]))

    def getNumFrames(self):
        if self._recordings is None:
            self._initialize()
        return self._recordings.shape[1]

    def getSamplingFrequency(self):
        if self._fs is None:
            self._initialize()
        return self._fs

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if self._recordings is None:
            self._initialize()
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = range(self.getNumChannels())
        return self._recordings[channel_ids, start_frame:end_frame]

    @staticmethod
    def writeRecording(recording, save_path):
        save_path = Path(save_path)
        if save_path.suffix == '.h5' or save_path.suffix == '.hdf5':
            F = h5py.File(save_path, 'w')
            info = {'recordings': {'fs': recording.getSamplingFrequency()}}
            F.create_dataset('info', data=json.dumps(info))
            if 'location' in recording.getChannelPropertyNames():
                positions = np.array([recording.getChannelProperty(chan, 'location')
                                      for chan in range(recording.getNumChannels())])
                F.create_dataset('channel_positions', data=positions)
            F.create_dataset('recordings', data=recording.getTraces())
            F.create_dataset('times', data=np.arange(recording.getNumFrames() / recording.getSamplingFrequency()))
            F.close()
            print('\nSaved recordings in', save_path, '\n')
        elif save_path is not None:
            save_folder = save_path
            if not save_folder.is_dir():
                save_folder.mkdir()
            np.save(save_folder / 'recordings', recording.getTraces())
            np.save(save_folder / 'timestamps', np.arange(recording.getNumFrames() / recording.getSamplingFrequency()))
            if 'location' in recording.getChannelPropertyNames():
                positions = np.array([recording.getChannelProperty(chan, 'location')
                                      for chan in range(recording.getNumChannels())])
                np.save(save_folder / 'channel_positions', positions)
            info = {'recordings': {'fs': recording.getSamplingFrequency()}}
            with (save_folder /'info.yaml').open('w') as f:
                yaml.dump(info, f, default_flow_style=False)
            print('\nSaved recordings in', save_folder, ' folder\n')


class MEArecSortingExtractor(SortingExtractor):
    def __init__(self, recording_path=None):
        SortingExtractor.__init__(self)
        self._recording_path = recording_path
        self._num_units = None
        self._spike_trains = None
        self._unit_ids = None
        self._fs = None
        self._initialize()

    def _initialize(self):
        rec_dict, info = load_recordings(recordings=self._recording_path)

        self._num_units = len(rec_dict['spiketrains'])
        if 'unit_id' in rec_dict['spiketrains'][0].annotations:
            self._unit_ids = [st.annotations['unit_id'] for st in rec_dict['spiketrains']]
        else:
            self._unit_ids = list(range(self._num_units))
        self._spike_trains = rec_dict['spiketrains']
        self._fs = info['recordings']['fs'] * pq.Hz  # fs is in kHz

    def getUnitIds(self):
        if self._unit_ids is None:
            self._initialize()
        return self._unit_ids

    def getNumUnits(self):
        if self._num_units is None:
            self._initialize()
        return self._num_units

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        if self._spike_trains is None:
            self._initialize()
        times = (self._spike_trains[self.getUnitIds().index(unit_id)].times.rescale('s') *
                 self._fs.rescale('Hz')).magnitude
        inds = np.where((start_frame <= times) & (times < end_frame))
        return np.rint(times[inds]).astype(int)

    @staticmethod
    def writeSorting(sorting, save_path, sampling_frequency):
        save_path = Path(save_path)
        if save_path.suffix == '.h5' or save_path.suffix == '.hdf5':
            F = h5py.File(save_path, 'w')
            info = {'recordings': {'fs': sampling_frequency, 'n_neurons': len(sorting.getUnitIds())}}
            F.create_dataset('info', data=json.dumps(info))
            for ii, unit in enumerate(sorting.getUnitIds()):
                st = sorting.getUnitSpikeTrain(unit) / float(sampling_frequency) * pq.s
                t_stop = np.max(sorting.getUnitSpikeTrain(unit)) / float(sampling_frequency) * pq.s
                F.create_dataset('spiketrains/{}/times'.format(unit), data=st)
                F.create_dataset('spiketrains/{}/t_stop'.format(unit), data=t_stop)
                annotations = {'unit_id': str(unit)}
                annotations_str = json.dumps(annotations)
                F.create_dataset('spiketrains/{}/annotations'.format(ii), data=annotations_str)
            F.close()
            print('\nSaved sorting in', save_path, '\n')
        elif save_path is not None:
            save_folder = save_path
            if not save_folder.is_dir():
                save_folder.mkdir()
            spiketrains = []
            for ii, unit in enumerate(sorting.getUnitIds()):
                st = sorting.getUnitSpikeTrain(unit) / sampling_frequency * pq.s
                t_stop = np.max(sorting.getUnitSpikeTrain(unit)) / sampling_frequency * pq.s
                spiketrain = neo.core.SpikeTrain(times=st, t_start=0 * pq.s, t_stop=t_stop)
                spiketrain.annotate(unit_id=unit)
                spiketrains.append(spiketrain)
            info = {'recordings': {'fs': sampling_frequency, 'n_neurons': len(spiketrains)}}
            np.save(save_folder / 'spiketrains', spiketrains)
            with (save_folder / 'info.yaml').open('w') as f:
                yaml.dump(info, f, default_flow_style=False)
            print('\nSaved sorting in', save_folder, ' folder\n')


def load_recordings(recordings, verbose=False):
    '''
    Load generated recordings (from template_gen.py)

    Parameters
    ----------
    recording_folder: recordings folder

    Returns
    -------
    recordings, times, positions, templates, spiketrains, sources, peaks - np.arrays
    info - dict

    '''
    if verbose:
        print("Loading recordings...")

    rec_dict = {}
    info = {}
    recordings = Path(recordings)
    if recordings.is_dir():
        recording_folder = recordings
        if (recording_folder /'recordings.npy').is_file():
            recordings = np.load(recording_folder /'recordings.npy')
            rec_dict.update({'recordings': recordings})
        if (recording_folder / 'channel_positions.npy').is_file():
            channel_positions = np.load(recording_folder / 'channel_positions.npy')
            rec_dict.update({'channel_positions': channel_positions})
        if (recording_folder / 'timestamps.npy').is_file():
            timestamps = np.load(recording_folder / 'timestamps.npy')
            rec_dict.update({'timestamps': timestamps})
        if (recording_folder / 'templates.npy').is_file():
            templates = np.load(recording_folder / 'templates.npy')
            rec_dict.update({'templates': templates})
        if (recording_folder / 'spiketrains.npy').is_file():
            spiketrains = np.load(recording_folder / 'spiketrains.npy')
            rec_dict.update({'spiketrains': spiketrains})
        if (recording_folder / 'spike_traces.npy').is_file():
            spike_traces = np.load(recording_folder / 'spike_traces.npy')
            rec_dict.update({'spike_traces': spike_traces})
        if (recording_folder / 'voltage_peaks.npy').is_file():
            voltage_peaks = np.load(recording_folder / 'voltage_peaks.npy')
            rec_dict.update({'voltage_peaks': voltage_peaks})
        with (recording_folder / 'info.yaml').open('r') as f:
            info = yaml.load(f)
    elif recordings.suffix == '.h5' or recordings.suffix == '.hdf5':
        with h5py.File(recordings, 'r') as F:
            info = json.loads(str(F['info'][()]))
            rec_dict['voltage_peaks'] = np.array(F.get('voltage_peaks'))
            rec_dict['channel_positions'] = np.array(F.get('channel_positions'))
            rec_dict['recordings'] = np.array(F.get('recordings'))
            rec_dict['spike_traces'] = np.array(F.get('spike_traces'))
            rec_dict['templates'] = np.array(F.get('templates'))
            rec_dict['timestamps'] = np.array(F.get('timestamps'))
            spiketrains = []
            if 'n_neurons' in info['recordings']:
                for ii in range(info['recordings']['n_neurons']):
                    times = np.array(F.get('spiketrains/{}/times'.format(ii))) * pq.s
                    t_stop = np.array(F.get('spiketrains/{}/t_stop'.format(ii))) * pq.s
                    annotations_str = str(F.get('spiketrains/{}/annotations'.format(ii))[()])
                    annotations = json.loads(annotations_str)
                    st = neo.core.SpikeTrain(
                        times,
                        t_stop=t_stop,
                        units=pq.s
                    )
                    st.annotations = annotations
                    spiketrains.append(st)
                rec_dict['spiketrains'] = spiketrains
    else:
        raise Exception("The provided file-folder is not a MEArec recording")

    if verbose:
        print("Done loading recordings...")

    return rec_dict, info



from spikeextractors import RecordingExtractor

import numpy as np
import h5py
import ctypes


class BiocamRecordingExtractor(RecordingExtractor):
    def __init__(self, recording_file):
        RecordingExtractor.__init__(self)
        self._recording_file = recording_file
        self._rf, self._nFrames, self._samplingRate, self._nRecCh, self._chIndices, self._file_format, self._signalInv, self._positions, self._read_function = openBiocamFile(
            self._recording_file)
        for m in range(self._nRecCh):
            self.setChannelProperty(m, 'location', self._positions[m])

    def getChannelIds(self):
        return list(range(self._nRecCh))

    def getNumFrames(self):
        return self._nFrames

    def getSamplingFrequency(self):
        return self._samplingRate

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = range(self.getNumChannels())
        data = self._read_function(
            self._rf, start_frame, end_frame, self.getNumChannels())
        return data.reshape((end_frame - start_frame,
                             self.getNumChannels())).T[channel_ids]

    @staticmethod
    def writeRecording(recording, save_path):
        M = recording.getNumChannels()
        N = recording.getNumFrames()
        channel_ids = range(M)
        raw = recording.getTraces()
        if raw.dtype != int:
            raise Exception('Cannot write dataset in the format with non-int datatype:', raw.dtype)
        rf = h5py.File(save_path, 'w')
        # writing out in 100 format: Time x Channels
        g = rf.create_group('3BData')
        d = rf.create_dataset('3BData/Raw', data=raw.T + 2048, dtype=int)
        g.attrs['Version'] = 100
        rf.create_dataset('3BRecInfo/3BRecVars/MinVolt', data=[0])
        rf.create_dataset('3BRecInfo/3BRecVars/MaxVolt', data=[1])
        rf.create_dataset('3BRecInfo/3BRecVars/NRecFrames', data=[N])
        rf.create_dataset('3BRecInfo/3BRecVars/SamplingRate', data=[recording.getSamplingFrequency()])
        rf.create_dataset('3BRecInfo/3BRecVars/SignalInversion', data=[1])
        rf.create_dataset('3BRecInfo/3BMeaChip/NCols', data=[M])
        rf.create_dataset('3BRecInfo/3BMeaStreams/Raw/Chs', data=np.vstack((np.arange(M), np.zeros(M))).T, dtype=int)
        rf.close()


def openBiocamFile(filename):
    """Open a Biocam hdf5 file, read and return the recording info, pick te correct method to access raw data, and return this to the caller."""
    rf = h5py.File(filename, 'r')
    # Read recording variables
    recVars = rf.require_group('3BRecInfo/3BRecVars/')
    # bitDepth = recVars['BitDepth'].value[0]
    # maxV = recVars['MaxVolt'].value[0]
    # minV = recVars['MinVolt'].value[0]
    nFrames = recVars['NRecFrames'].value[0]
    samplingRate = recVars['SamplingRate'].value[0]
    signalInv = recVars['SignalInversion'].value[0]

    # Read chip variables
    chipVars = rf.require_group('3BRecInfo/3BMeaChip/')
    nCols = chipVars['NCols'].value[0]

    # Get the actual number of channels used in the recording
    file_format = rf['3BData'].attrs.get('Version')
    if file_format == 100:
        nRecCh = len(rf['3BData/Raw'][0])
        # raise Warning('This may go wrong!')
    elif file_format == 101:
        nRecCh = int(1. * rf['3BData/Raw'].shape[0] / nFrames)
    else:
        raise Exception('Unknown data file format.')

    print('# 3Brain data format:', file_format, 'signal inversion', signalInv)
    print('#       signal range: ', recVars['MinVolt'].value[0], '- ',
          recVars['MaxVolt'].value[0])
    # Compute indices
    rawIndices = rf['3BRecInfo/3BMeaStreams/Raw/Chs'].value

    # Name channels ([0..4095] for fullarray files)
    chIndices = [(x - 1) + (y - 1) * nCols for (y, x) in rawIndices]
    # chIndices = [(x-1) + (y-1)*nCols for (x,y) in rawIndices]
    # Swap X and Y (old format)

    # determine correct function to read data
    print("# Signal inversion looks like " + str(signalInv) + ", guessing the "
                                                              "right method for data access.\n# If your results "
                                                              "look strange, signal polarity is wrong.")
    if file_format == 100:
        if signalInv == -1:
            read_function = readHDF5t_100
        else:
            read_function = readHDF5t_100_i
    else:
        if signalInv == -1:
            read_function = readHDF5t_101_i
        else:
            read_function = readHDF5t_101

    return (rf, nFrames, samplingRate, nRecCh, chIndices, file_format, signalInv, rawIndices, read_function)


def readHDF5(rf, t0, t1):
    """In order to use the algorithms designed for the old format, the input data must be inverted."""
    return 4095 - rf['3BData/Raw'][t0:t1].flatten().astype(ctypes.c_short)


def readHDF5t_100(rf, t0, t1, nch):
    """Transposed version for the interpolation method."""
    if t0 <= t1:
        d = 2048 - rf['3BData/Raw'][t0:t1].flatten('C').astype(ctypes.c_short)
        d[np.where(np.abs(d) > 1500)[0]] = 0
        return d
    else:  # Reversed read
        raise Exception('Reading backwards? Not sure about this.')
        return 2048 - rf['3BData/Raw'][t1:t0].flatten(
            'F').astype(ctypes.c_short)


def readHDF5t_100_i(rf, t0, t1, nch):
    ''' Transposed version for the interpolation method. '''
    if t0 <= t1:
        d = rf['3BData/Raw'][t0:t1].flatten('C').astype(ctypes.c_short) - 2048
        d[np.where(np.abs(d) > 1500)[0]] = 0
        return d
    else:  # Reversed read
        raise Exception('Reading backwards? Not sure about this.')
        return rf['3BData/Raw'][t1:t0].flatten(
            'F').astype(ctypes.c_short) - 2048


def readHDF5t_101(rf, t0, t1, nch):
    ''' Transposed version for the interpolation method. '''
    if t0 <= t1:
        d = rf['3BData/Raw'][nch * t0:nch * t1].reshape(
            (-1, nch), order='C').flatten('C').astype(ctypes.c_short) - 2048
        d[np.abs(d) > 1500] = 0
        return d
    else:  # Reversed read
        raise Exception('Reading backwards? Not sure about this.')
        d = rf['3BData/Raw'][nch * t1:nch * t0].reshape(
            (-1, nch), order='C').flatten('C').astype(ctypes.c_short) - 2048
        d[np.where(np.abs(d) > 1500)[0]] = 0
        return d


def readHDF5t_101_i(rf, t0, t1, nch):
    ''' Transposed version for the interpolation method. '''
    if t0 <= t1:
        d = 2048 - rf['3BData/Raw'][nch * t0:nch * t1].reshape(
            (-1, nch), order='C').flatten('C').astype(ctypes.c_short)
        d[np.where(np.abs(d) > 1500)[0]] = 0
        return d
    else:  # Reversed read
        raise Exception('Reading backwards? Not sure about this.')
        d = 2048 - rf['3BData/Raw'][nch * t1:nch * t0].reshape(
            (-1, nch), order='C').flatten('C').astype(ctypes.c_short)
        d[np.where(np.abs(d) > 1500)[0]] = 0
        return d



from spikeextractors import SortingExtractor

import numpy as np
import h5py


class HS2SortingExtractor(SortingExtractor):
    def __init__(self, recording_file):
        SortingExtractor.__init__(self)
        self._recording_file = recording_file
        self._rf = h5py.File(self._recording_file, mode='r')
        self._unit_ids = set(self._rf['cluster_id'].value)
        if 'centres' in self._rf.keys():
            self._unit_locs = self._rf['centres'].value  # cache for faster access

    def get_unit_indices(self, x):
        return np.where(self._rf['cluster_id'].value == x)[0]

    def getUnitIds(self):
        return list(self._unit_ids)

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        times = self._rf['times'].value[self.get_unit_indices(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    def getUnitChannels(self, unit_id):
        return self._rf['ch'].value[self.get_unit_indices(unit_id)]

    def getUnitProperty(self, unit_id, property):
        if property == 'spike_locations':
            return self._rf['data'][:2, self.get_unit_indices(unit_id)].T
        elif property == 'unit_location':
            return self._unit_locs[unit_id]
        else:
            raise Exception("Property does not exist.")

    @staticmethod
    def writeSorting(sorting, save_path):
        unit_ids = sorting.getUnitIds()
        times_list = []
        labels_list = []
        for i in range(len(unit_ids)):
            unit = unit_ids[i]
            times = sorting.getUnitSpikeTrain(unit_id=unit)
            times_list.append(times)
            labels_list.append(np.ones(times.shape, dtype=int) * unit)
        all_times = np.concatenate(times_list)
        all_labels = np.concatenate(labels_list)
        rf = h5py.File(save_path, mode='w')
        # for now only create the entries required by any RecordingExtractor
        rf.create_dataset("times", data=all_times)
        rf.create_dataset("cluster_id", data=all_labels)
        rf.close()



from .SortingExtractor import SortingExtractor
import numpy as np


# A Sorting Extractor that allows for manual curation of a sorting result (Represents curation as a tree of units)

class CuratedSortingExtractor(SortingExtractor):

    def __init__(self, parent_sorting):
        SortingExtractor.__init__(self)
        self._parent_sorting = parent_sorting
        self._original_unit_ids = list(np.copy(parent_sorting.getUnitIds()))
        self._all_ids = list(np.copy(parent_sorting.getUnitIds()))
        original_spike_indices = []
        for unit_id in self._original_unit_ids:
            original_spike_indices.append(np.asarray(list(range(len(parent_sorting.getUnitSpikeTrain(unit_id))))))
        self._indices_dict = {}
        for i, unit_id in enumerate(self._original_unit_ids):
            self._indices_dict[unit_id] = original_spike_indices[i]
        del original_spike_indices

        self._roots = []
        for i, unit_id in enumerate(self._original_unit_ids):
            root = Unit(unit_id)
            self._roots.append(root)

    def getUnitIds(self):
        unit_ids = []
        for root in self._roots:
            unit_ids.append(root.unit_id)
        return unit_ids

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        valid_unit_id = False
        spike_train = np.asarray([])
        for root in self._roots:
            if(root.unit_id == unit_id):
                valid_unit_id = True
                self._collectSpikeTrains(root)
                inds = np.where((start_frame <= root.spike_train) & (root.spike_train < end_frame))
                spike_train = root.spike_train[inds]
                self._resetSpikeTrains(root)
        if(valid_unit_id):
            return spike_train
        else:
            raise ValueError("invalid unit id")

    def _collectSpikeTrains(self, unit): #post order traversal
        for child in unit.get_children():
            self._collectSpikeTrains(child)
        if(len(unit.get_children()) == 0):
            unit.spike_train = self._parent_sorting.getUnitSpikeTrain(unit.unit_id)
        else:
            new_spike_train = []
            for child in unit.get_children():
                new_spike_train.append(child.spike_train)
            new_spike_train = np.asarray(new_spike_train)
            new_spike_train = np.sort(np.concatenate(new_spike_train).ravel())
            new_spike_train = new_spike_train[self._indices_dict[unit.unit_id]]
            unit.spike_train = new_spike_train

    def _resetSpikeTrains(self, unit): #post order traversal
        for child in unit.get_children():
            self._resetSpikeTrains(child)
        unit.spike_train = np.asarray([])

    def printCurationTree(self, unit_id):
        '''This function prints the current curation tree for the unit_id (roots are current unit ids).

        Parameters
        ----------
        unit_id: in
            The unit id whose curation history will be printed.
        '''
        root_ids = []
        for i in range(len(self._roots)):
            root_id = self._roots[i].unit_id
            root_ids.append(root_id)
        if(unit_id in root_ids):
            root_index = root_ids.index(unit_id)
            print(self._roots[root_index])
        else:
            raise ValueError("invalid unit id")


    def excludeUnits(self, unit_ids):
        '''This function deletes roots from the curation tree according to the given unit_ids

        Parameters
        ----------
        unit_ids: list
            The unit ids to be excluded
        '''
        root_ids = []
        for i in range(len(self._roots)):
            root_id = self._roots[i].unit_id
            root_ids.append(root_id)

        if(set(unit_ids).issubset(set(root_ids)) and len(unit_ids) > 0):
            indices_to_be_deleted = []
            for unit_id in unit_ids:
                root_index = root_ids.index(unit_id)
                indices_to_be_deleted.append(root_index)
                del self._indices_dict[unit_id]
            self._roots = [self._roots[i] for i,_ in enumerate(root_ids) if i not in indices_to_be_deleted]
        else:
            raise ValueError("invalid unit ids")

    def mergeUnits(self, unit_ids):
        '''This function merges two roots from the curation tree according to the given unit_ids. It creates a new unit_id and root
        that has the merged roots as children.

        Parameters
        ----------
        unit_ids: list
            The unit ids to be merged
        '''
        root_ids = []
        for i in range(len(self._roots)):
            root_id = self._roots[i].unit_id
            root_ids.append(root_id)
        if(set(unit_ids).issubset(set(root_ids)) and len(unit_ids) > 1):
            new_root_id = max(self._all_ids)+1
            self._all_ids.append(new_root_id)
            new_root = Unit(new_root_id)
            indices_to_be_deleted = []
            self._indices_dict[new_root_id] = []
            for unit_id in unit_ids:
                root_index = root_ids.index(unit_id)
                new_root.add_child(self._roots[root_index])
                for spike_index in self._indices_dict[unit_id]:
                    self._indices_dict[new_root_id].append(len(self._indices_dict[new_root_id]))
                indices_to_be_deleted.append(root_index)
            self._indices_dict[new_root_id] = np.asarray(self._indices_dict[new_root_id])
            self._roots = [self._roots[i] for i,_ in enumerate(root_ids) if i not in indices_to_be_deleted]
            self._roots.append(new_root)
        else:
            raise ValueError("invalid unit ids")

    def splitUnit(self, unit_id, indices):
        '''This function splits a root from the curation tree according to the given unit_id and indices. It creates two new unit_ids
        and roots that have the split root as a child. This function splits the spike train of the root by the given indices.

        Parameters
        ----------
        unit_id: int
            The unit id to be split
        indices: list
            The indices of the unit spike train at which the spike train will be split.
        '''
        root_ids = []
        for i in range(len(self._roots)):
            root_id = self._roots[i].unit_id
            root_ids.append(root_id)

        if(unit_id in root_ids):
            root_index = root_ids.index(unit_id)
            root_spike_indices = self._indices_dict[unit_id]
            if(set(indices).issubset(set(root_spike_indices)) and len(indices) > 0 and len(indices) < len(root_spike_indices)):
                new_child = self._roots[root_index]

                new_root_1_id = max(self._all_ids)+1
                self._all_ids.append(new_root_1_id)
                new_root_1 = Unit(new_root_1_id)
                new_root_1.add_child(new_child)
                self._indices_dict[new_root_1_id] = np.asarray(indices)

                non_indices = [index for i, index in enumerate(root_spike_indices) if i not in indices]
                new_root_2_id = max(self._all_ids)+1
                self._all_ids.append(new_root_2_id)
                new_root_2 = Unit(new_root_2_id)
                new_root_2.add_child(new_child)
                self._indices_dict[new_root_2_id] = np.asarray(non_indices)

                del self._roots[root_index]
                self._roots.append(new_root_1)
                self._roots.append(new_root_2)
            else:
                raise ValueError("invalid indices")
        else:
            raise ValueError("invalid unit ids")


# The Unit class is a node in the curation tree. Each Unit contains its unit_id, children, and spike_train.
class Unit(object):
    def __init__(self, unit_id):
        self.unit_id = unit_id
        self.children = []
        self.spike_train = np.asarray([])

    def add_child(self, child):
        self.children.append(child)

    def get_children(self):
        return self.children

    def __str__(self, level=0):
        if(level == 0):
            ret = "\t"*(max(level-1, 0)) +repr(self.unit_id)+ "\n"
        else:
            ret = "\t"*(max(level-1, 0)) + "^-------" +repr(self.unit_id)+ "\n"
        for child in self.children:
            ret += child.__str__(level+1)
        return ret



from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor

import json
import numpy as np
from pathlib import Path


def _load_required_modules():
    try:
        from mountainlab_pytools import mdaio
        from kbucket import client as kbucket
    except ModuleNotFoundError:
        raise ModuleNotFoundError("To use the MdaExtractors, install mountainlab_pytools and kbucket: \n\n"
                                  "pip install mountainlab_pytools kbucket\n\n")
    return mdaio, kbucket


class MdaRecordingExtractor(RecordingExtractor):
    def __init__(self, dataset_directory, *, download=True):
        mdaio, kbucket = _load_required_modules()

        RecordingExtractor.__init__(self)
        dataset_directory = Path(dataset_directory)
        self._dataset_directory = dataset_directory
        timeseries0 = dataset_directory / 'raw.mda'
        self._dataset_params = read_dataset_params(dataset_directory)
        self._samplerate = self._dataset_params['samplerate'] * 1.0
        if is_kbucket_url(str(timeseries0)):
            download_needed = is_url(kbucket.findFile(str(timeseries0)))
        else:
            download_needed = is_url(str(timeseries0))
        if download and download_needed:
            print('Downloading file: ' + timeseries0)
            self._timeseries_path = kbucket.realizeFile(str(timeseries0))
            print('Done.')
        else:
            self._timeseries_path = kbucket.findFile(str(timeseries0))
        geom0 = dataset_directory / 'geom.csv'
        self._geom_fname = kbucket.realizeFile(str(geom0))
        self._geom = np.genfromtxt(self._geom_fname, delimiter=',')
        X = mdaio.DiskReadMda(self._timeseries_path)
        if self._geom.shape[0] != X.N1():
            raise Exception(
                'Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0],
                                                                                               X.N1()))
        self._num_channels = X.N1()
        self._num_timepoints = X.N2()
        for m in range(self._num_channels):
            self.setChannelProperty(m, 'location', self._geom[m, :])

    def getChannelIds(self):
        return list(range(self._num_channels))

    def getNumFrames(self):
        return self._num_timepoints

    def getSamplingFrequency(self):
        return self._samplerate

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        mdaio, kbucket = _load_required_modules()
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = self.getChannelIds()
        X = mdaio.DiskReadMda(self._timeseries_path)
        recordings = X.readChunk(i1=0, i2=start_frame, N1=X.N1(), N2=end_frame - start_frame)
        recordings = recordings[channel_ids, :]
        return recordings

    @staticmethod
    def writeRecording(recording, save_path):
        mdaio, kbucket = _load_required_modules()
        save_path = Path(save_path)
        channel_ids = recording.getChannelIds()
        M = len(channel_ids)
        N = recording.getNumFrames()
        raw = recording.getTraces()

        if not save_path.is_dir():
            save_path.mkdir()

        if 'location' in recording.getChannelPropertyNames():
            location0 = recording.getChannelProperty(channel_ids[0], 'location')
            nd = len(location0)
            geom = np.zeros((M, nd))
            for ii in range(len(channel_ids)):
                location_ii = recording.getChannelProperty(channel_ids[ii], 'location')
                geom[ii, :] = list(location_ii)
            np.savetxt(save_path / 'geom.csv', geom, delimiter=',')

        mdaio.writemda32(raw, str(save_path / 'raw.mda'))
        params = dict(
            samplerate=recording.getSamplingFrequency(),
            spike_sign=-1
        )
        with (save_path / 'params.json').open('w') as f:
            json.dump(params, f)


class MdaSortingExtractor(SortingExtractor):
    def __init__(self, firings_file):
        mdaio, kbucket = _load_required_modules()

        SortingExtractor.__init__(self)
        if is_kbucket_url(firings_file):
            download_needed = is_url(kbucket.findFile(firings_file))
        else:
            download_needed = is_url(firings_file)
        if download_needed:
            print('Downloading file: ' + firings_file)
            self._firings_path = kbucket.realizeFile(firings_file)
            print('Done.')
        else:
            self._firings_path = kbucket.realizeFile(firings_file)
        self._firings = mdaio.readmda(self._firings_path)
        self._times = self._firings[1, :]
        self._labels = self._firings[2, :]
        self._unit_ids = np.unique(self._labels).astype(int)

    def getUnitIds(self):
        return self._unit_ids

    def getUnitSpikeTrain(self, unit_id, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        inds = np.where((self._labels == unit_id) & (start_frame <= self._times) & (self._times < end_frame))
        return np.rint(self._times[inds]).astype(int)

    @staticmethod
    def writeSorting(sorting, save_path):
        mdaio, kbucket = _load_required_modules()
        save_path = Path(save_path)
        unit_ids = sorting.getUnitIds()
        if len(unit_ids) > 0:
            K = np.max(unit_ids)
        else:
            K = 0
        times_list = []
        labels_list = []
        for i in range(len(unit_ids)):
            unit = unit_ids[i]
            times = sorting.getUnitSpikeTrain(unit_id=unit)
            times_list.append(times)
            labels_list.append(np.ones(times.shape) * unit)
        all_times = _concatenate(times_list)
        all_labels = _concatenate(labels_list)
        sort_inds = np.argsort(all_times)
        all_times = all_times[sort_inds]
        all_labels = all_labels[sort_inds]
        L = len(all_times)
        firings = np.zeros((3, L))
        firings[1, :] = all_times
        firings[2, :] = all_labels
        mdaio.writemda64(firings, str(save_path))


def _concatenate(list):
    if len(list) == 0:
        return np.array([])
    return np.concatenate(list)


def is_kbucket_url(path):
    return path.startswith('kbucket://') or path.startswith('sha1://')


def is_url(path):
    return path.startswith('http://') or path.startswith('https://') or path.startswith(
        'kbucket://') or path.startswith('sha1://')


def read_dataset_params(dsdir):
    mdaio, kbucket = _load_required_modules()

    params_fname = Path(kbucket.realizeFile(str(dsdir / 'params.json')))
    if not params_fname.is_file():
        raise Exception('Dataset parameter file does not exist: ' + str(params_fname))
    with params_fname.open() as f:
        return json.load(f)



from spikeextractors import RecordingExtractor
import os
import numpy as np
from pathlib import Path


class BinDatRecordingExtractor(RecordingExtractor):
    def __init__(self, datfile, samplerate, numchan, dtype=None, geom=None):
        RecordingExtractor.__init__(self)
        self._datfile = Path(datfile)
        if dtype == None:
            dtype = np.float32
        self._timeseries = _read_binary(self._datfile, numchan, dtype)
        self._samplerate = float(samplerate)
        self._geom = geom
        if geom is not None:
            for m in range(self._timeseries.shape[0]):
                self.setChannelProperty(m, 'location', self._geom[m, :])

    def getChannelIds(self):
        return list(range(self._timeseries.shape[0]))

    def getNumFrames(self):
        return self._timeseries.shape[1]

    def getSamplingFrequency(self):
        return self._samplerate

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.getNumFrames()
        if channel_ids is None:
            channel_ids = self.getChannelIds()
        recordings = self._timeseries[:, start_frame:end_frame][channel_ids, :]
        return recordings

    @staticmethod
    def writeRecording(recording, save_path, dtype=None, transpose=False):
        save_path = Path(save_path)
        if dtype == None:
            dtype = np.float32
        if not transpose:
            if not save_path.suffix == '.dat':
                save_path = save_path.parent / (save_path.name + '.dat')
            with save_path.open('wb') as f:
                np.transpose(np.array(recording.getTraces(), dtype=dtype)).tofile(f)
        elif transpose:
            if not save_path.suffix == '.dat':
                save_path = save_path.parent / (save_path.name + '.dat')
            with save_path.open('wb') as f:
                np.array(recording.getTraces(), dtype=dtype).tofile(f)


def _read_binary(file, numchan, dtype):
    numchan = int(numchan)
    assert file.suffix == '.dat'
    with Path(file).open() as f:
        nsamples = os.fstat(f.fileno()).st_size // (numchan * np.dtype(dtype).itemsize)
        samples = np.memmap(f, np.dtype(dtype), mode='r',
                            shape=(nsamples, numchan))
        samples = np.transpose(samples)
    return samples
